@@ -34,8 +34,14 @@ import av
 
 
 class CallbackPopulation(BasicPopulation):
-    def __init__(self, field_dict, population_size, automata_class, **kwargs):
-        super(CallbackPopulation, self).__init__(field_dict, population_size, automata_class, **kwargs)
+    def __init__(self, *args, **kwargs):
+        """The CallbackPopulation instance initializer expects the same parameters as
+        the BasicPopulation and does the same.
+
+        :param args: positional parameters
+        :param kwargs: keyword parameters
+        """
+        super(CallbackPopulation, self).__init__(*args, **kwargs)
 
         self.callbacks = []
 
@@ -47,9 +53,13 @@ class CallbackPopulation(BasicPopulation):
         """
         self.callbacks.append(callback_instance)
 
-    def play(self, stop_after=None, **kwargs):
-        stop_after = stop_after if stop_after is not None else self.stop_after
+    def play(self, stop_after=1):
+        """From the automata perspective the play() method behaves the same way as
+        the BasicPopulation.play() method. However, it also call all the registered callbacks
+        after every generation.
 
+        :param stop_after: int
+        """
         start_time = time.time()
         for _ in range(stop_after):
             if self.run() != self.Status.RUNNABLE:
@@ -62,25 +72,70 @@ class CallbackPopulation(BasicPopulation):
 
 
 class Callback(ABC):
+    """The Callback abstract class """
     def __init__(self, population, base_directory=None, activation_condition_function=(lambda p: True)):
+        """The Callback abstract class initializer method prepares some common properties
+        of the subclasses. It must be passed a reference to the monitored population and has
+        two optional keyword parameters.
+
+        base_directory allows to specify a destination directory for all the files (if any)
+        produced by the monitoring process. By default, the destination will be a timestamped
+        subdirectory of the current directory.
+
+        activation_condition_function is a function controlling how often or when the callbacks
+        will be activated. By default, the registered callback are activated at each generation.
+        The activation_condition_function is a function that takes a population and returns
+        True if the current population state will have to be reported (printed, saved, added to
+        a video, ...) In most cases, the result will depend on the population generation.
+
+        To minimize coding, there are three predefined functions in the class:
+
+        - condition_each_n_generation()
+
+        - condition_at_generation()
+
+        - condition_or()
+
+        :param population: CallbackPopulation
+        :param base_directory: str
+        :param activation_condition_function: a function
+        """
         self.population = population
         self.activation_condition_function = activation_condition_function
 
         if base_directory is None:
-            base_directory = self.population.automata_class.__name__ # describe()
+            base_directory = self.population.automata_class.__name__  # or describe() ?
         # Make the base directory name unique by adding the date and PID
         self.base_directory = f"{base_directory}_{datetime.datetime.now().strftime('%Y%m%d')}_{os.getpid()}"
 
     @staticmethod
     def condition_each_n_generation(report_progress_each):
+        """The condition_each_n_generation() static method is a function generator. The produced function
+        will expect a population as parameter and will return True if the generation number of the
+        population is divisible by report_progress_each, the parameter passed to the function generator.
+
+        :param report_progress_each: int
+        """
         return lambda population: (population.generation % report_progress_each) == 0
 
     @staticmethod
     def condition_at_generation(generation):
+        """The condition_at_generation() static method is a function generator. The produced function
+        will expect a population as parameter and will return True if the generation number of the
+        population equals generation, the parameter passed to the function generator.
+
+        :param generation: int
+        """
         return lambda population: population.generation == generation
 
     @staticmethod
     def condition_or(activation_condition_function_list):
+        """The condition_or() static method is a function generator. The produced function
+        will expect a population as parameter and will return True if (at least) one of the
+        activation function returns True.
+
+        :param activation_condition_function_list: activation function list
+        """
         return lambda population: any([activation_condition_function(population)
                                        for activation_condition_function in activation_condition_function_list])
 
@@ -93,9 +148,31 @@ class Callback(ABC):
 
 class LogProgressCallback(Callback):
     def __init__(self, *args, logfile=None, loglevel=logging.INFO, **kwargs):
+        """The initializer method prepares the logging. It takes two specific keyword parameters to
+        control the destination and the desired level of logging.
+
+        The logfile parameter may receive a string denoting the path of the logfile. It is the responsibility
+        of the caller to ensure the directory which will receive the logfile exists.
+        The logfile parameter may also set to True, in this case the logfile will have a default location
+        depending on the base_directory keyword parameter.
+        If logfile parameter is None (by default), the logging will be reported to the console.
+
+        The loglevel parameter sets the expected level of logging.
+        See: https://docs.python.org/3/library/logging.html#levels
+
+        :param logfile: str, bool or None
+        :param loglevel: int (logging level)
+        """
         super(LogProgressCallback, self).__init__(*args, **kwargs)
 
         if logfile:
+            # If the logfile parameter value is a boolean, it means the logfile will
+            # have a default place
+            if isinstance(logfile, bool):
+                if not os.path.exists(self.base_directory):
+                    os.makedirs(self.base_directory)
+                logfile = os.path.join(self.base_directory, "progress.log")
+
             logging.basicConfig(level=loglevel, filename=logfile)
         else:
             logging.basicConfig(level=loglevel)
@@ -105,8 +182,9 @@ class LogProgressCallback(Callback):
     def callback(self, elapsed_time):
         if self.activation_condition_function(self.population):
             # Some progress has been done, log it
-            logging.info(f"generation = {self.population.generation} "
-                         f"(elapsed time {elapsed_time:.3f}s, average {self.population.generation / elapsed_time:.3f}gps)")
+            logging.info(f"{self.population.describe()} "
+                         f"(elapsed time {elapsed_time:.3f}s,"
+                         f" average {self.population.generation / elapsed_time:.3f}gps)")
 
 
 class ImageCallback(Callback):
@@ -133,17 +211,18 @@ class ImageCallback(Callback):
         return f"{str(self.population.generation).zfill(self.zero_padding_length)}"
 
 
-class SaveTracesImageCallback(ImageCallback):
-    class Trace(IntFlag):
-        TRAJECTORIES = 0b01
-        POSITIONS = 0b10
+class Trace(IntFlag):
+    TRAJECTORIES = 0b01
+    POSITIONS = 0b10
 
+
+class SaveTracesImageCallback(ImageCallback):
     def __init__(self, *args, trace=Trace.TRAJECTORIES+Trace.POSITIONS, **kwargs):
         super(SaveTracesImageCallback, self).__init__(*args, **kwargs)
 
         field_size = self._get_field_size()
 
-        if trace & self.Trace.TRAJECTORIES:
+        if trace & Trace.TRAJECTORIES:
             # Create a directory to save the automata trajectories
             self.trajectories_save_path = os.path.join(self.base_directory, f"trajectories")
             if not os.path.exists(self.trajectories_save_path):
@@ -153,7 +232,7 @@ class SaveTracesImageCallback(ImageCallback):
         else:
             self.trajectories_monitor = None
 
-        if trace & self.Trace.POSITIONS:
+        if trace & Trace.POSITIONS:
             # Create a directory to save the automata position
             self.positions_save_path = os.path.join(self.base_directory, f"positions")
             if not os.path.exists(self.positions_save_path):
@@ -307,10 +386,6 @@ class VideoCallback(Callback):
 
 
 class SaveTracesVideoCallback(VideoCallback):
-    class Trace(IntFlag):
-        TRAJECTORIES = 0b01
-        POSITIONS = 0b10
-
     def __init__(self, *args, trace=Trace.TRAJECTORIES+Trace.POSITIONS, **kwargs):
         super(SaveTracesVideoCallback, self).__init__(*args, **kwargs)
 
@@ -324,7 +399,7 @@ class SaveTracesVideoCallback(VideoCallback):
                 field_size = subfield.size
                 break
 
-        if trace & self.Trace.TRAJECTORIES:
+        if trace & Trace.TRAJECTORIES:
             # Prepare a monitor to keep track of the automata trajectories
             self.trajectories_monitor = PILMonitor(field_size, ColorGradient.create())
 
@@ -335,7 +410,7 @@ class SaveTracesVideoCallback(VideoCallback):
             self.trajectories_monitor = None
             self.trajectories_avdata = None
 
-        if trace & self.Trace.POSITIONS:
+        if trace & Trace.POSITIONS:
             # Prepare a monitor to keep track of the automata positions
             self.positions_monitor = PILMonitor(field_size, ColorGradient.create(gradient_name='cgy'))
 
@@ -414,5 +489,3 @@ class SaveFieldsVideoCallback(VideoCallback):
     def __del__(self):
         for avdata in self.subfield_avdata.values():
             self.__class__._close_avdata(avdata)
-
-
